from inspect import Parameter
import itertools


DEFAULT_KINDS = [Parameter.POSITIONAL_ONLY, Parameter.POSITIONAL_OR_KEYWORD]


def accumulate(args: tuple, kwargs: dict, params, kinds: list = None):
    '''accumulate gathers and yields all args / kwargs that have annotations matching the given kinds list.

    kinds should be a list of Parameter.kind.
    eg: [Parameter.POSITIONAL_ONLY, Parameter.POSITIONAL_OR_KEYWORD]
    '''
    kinds = kinds or DEFAULT_KINDS
    items = { 
        param.name: args[index]
        for index, param in enumerate(params.values())
            if index < len(args)
            and param.kind in kinds
    }

    for name, value in itertools.chain(items.items(), kwargs.items()):
        param = params.get(name, None)

        if (param is None or param.default == value
        or param.annotation == param.empty):
            continue

        yield (name, value, param)



import inspect

from . import errors
from .accumulate import accumulate


def custom(equality):
    def check(func):

        sig = inspect.signature(func)
        prms = dict(sig.parameters.items())

        def decorate(*args, **kwargs):
            for name, value, param in accumulate(args, kwargs, prms):

                try:
                    eq = bool(equality(value, param.annotation))
                except Exception as err:
                    raise errors.TypeCheck(f"Param '{name}' failed in equality check for '{param.annotation}', given value '{value}'")

                if not eq:
                    raise errors.TypeCheck(f"Param '{name}' does not satisfy '{param.annotation}', given value '{value}'")

            return_value = func(*args, **kwargs)

            if sig.return_annotation == sig.empty:
                return return_value

            try:
                eq = bool(equality(return_value, sig.return_annotation))
            except Exception as err:
                raise errors.TypeCheck(f"Return value failed in equality check for '{sig.return_annotation}', given value '{return_value}'")

            if not eq:
                raise errors.TypeCheck(f"Return value does not satisfy '{sig.return_annotation}', given value '{return_value}'")

            return return_value

        return decorate
    return check


cast = custom(lambda value, annotation: annotation(value))
instance = custom(isinstance)



from setuptools import setup
import os


NAME = 'annotyped'
DESCRIPTION = 'A lightweight library for type checking and type casting annotated vars.'
URL = 'https://gitlab.com/sj1k/annotyped'
EMAIL = '42Echo6Alpha@gmail.com'
AUTHOR = 'Steven "sj1k" Core'
REQUIRES_PYTHON = '>=3.5.0'
VERSION = '0.2.4'

here = os.path.abspath(os.path.dirname(__file__))

try:
    with open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
        long_description = '\n' + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION


setup(
    name=NAME,
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/markdown",
    author=AUTHOR,
    email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    version=VERSION,
    packages=['annotyped'],
)



import inspect

from . import errors
from .accumulate import accumulate


def cast(func):

    # Inspect the signature and params for later lookup.
    sig = inspect.signature(func)
    prms = dict(sig.parameters.items())

    def decorate(*args, **kwargs):
        args = list(args)
        for index, (name, value, param) in enumerate(accumulate(args, kwargs, prms)):
            try:
                value = param.annotation(value)
            except Exception as err:
                raise errors.TypeCast(f"Param '{name}' could not convert to '{param.annotation}' from '{type(value)}': {err}") from err

            if index < len(args):
                args[index] = value
            else:
                kwargs[name] = value

        # Store the return value for now, we may need it.
        return_value = func(*args, **kwargs)

        if sig.return_annotation == sig.empty:
            return return_value

        try:
            return_value = sig.return_annotation(return_value)
        except Exception as err:
            raise errors.TypeCast(f"Return value could not convert to '{param.annotation}' from '{type(return_value)}': {err}") from err

        return return_value
    return decorate
